<?php
/**
 * Daisycon api connection
 *
 * @category  Daisycon
 * @package   Daisycon_Affiliatemarketing
 * @copyright Copyright (c) 2018 Daisycon B.V.
 * @author    Daisycon B.V. <support@daisycon.com>
 */
class Daisycon_Affiliatemarketing_Model_Api_Connection
{
    /**
     * The service base url
     *
     * @var string
     */
    protected $_baseUrl = '';

    /**
     * The location of th API
     *
     * @var string
     */
    protected $_apiUrl = '';

    /**
     * Retrieve a token
     *
     * @var string
     */
    protected $_authUrl = 'authenticate';

    /**
     * Username
     *
     * @var string|null
     */
    protected $_username;

    /**
     * User's password
     *
     * @var string|null
     */
    protected $_password;

    /**
     * Access token
     *
     * @var string
     */
    protected $_accessToken;

    /**
     * Http client
     *
     * @var Daisycon_Affiliatemarketing_Model_Api_Client
     */
    protected $_client;

    /**
     * Url placeholders
     *
     * @var array
     */
    protected $_placeholders = array();

    /**
     * Next url - for paging
     *
     * @var string
     */
    protected $_nextUrl = null;

    /**
     * Init connection
     * 
     * @param array $options
     */
    public function __construct(array $options = array())
    {
        if (true === isset($options['username'])) {
            $this->setUsername($options['username']);
        }

        if (true === isset($options['password'])) {
            $this->setPassword($options['password']);
        }

        $this->_baseUrl = $this->_getConfig()->getApiServiceUrl();
    }

    /**
     * Get config
     *
     * @return Daisycon_Affiliatemarketing_Model_Config
     */
    protected function _getConfig()
    {
        return Mage::getSingleton('affiliatemarketing/config');
    }

    /**
     * Get the http client
     *
     * @return Daisycon_Affiliatemarketing_Model_Api_Client
     */
    protected function _getClient()
    {
        if (null !== $this->_client) {
            return $this->_client;
        }

        $this->_client = new Daisycon_Affiliatemarketing_Model_Api_Client();

        return $this->_client;
    }

    /**
     * Add a url place holder
     *
     * @param string $placeholder
     * @param mixed value
     */
    public function addPlaceholderValue($placeholder, $value = null)
    {
        $this->_placeholders[$placeholder] = $value;
        return $this;
    }

    /**
     * Return the connected client, authorize if required
     *
     * @return Daisycon_Affiliatemarketing_Model_Api_Client
     */
    public function connect()
    {
        // If access token is not set or token has expired, acquire new token
        if (true === empty($this->accessToken)) {
            $this->_acquireAccessToken();
        }

        $client = $this->_getClient();

        return $client;
    }

    /**
     * Create an authenticated request object
     *
     * @param string $method
     * @param string $endPoint
     * @param string|null $body
     * @param array $headers
     * @return Daisycon_Affiliatemarketing_Model_Api_Request
     * @throws Daisycon_Affiliatemarketing_Model_Api_InvalidResponseException
     */
    protected function _createAuthenticatedRequest(
        $method = Daisycon_Affiliatemarketing_Model_Api_Request::GET,
        $endPoint = '',
        $body = null,
        array $headers = array()
    ) {

        // If access token is not set or token has expired, acquire new token
        if (true === empty($this->_accessToken)) {
            $this->_acquireAccessToken();
        }

        // If we have a token, sign the request
        if (false === $this->needsAuthentication() && false === empty($this->_accessToken)) {
            $headers['Authorization'] = 'Bearer ' . $this->_accessToken;
        }

        return $this->_createRequest(
            $method,
            $endPoint,
            $body,
            $headers
        );
    }

    /**
     * Create a request
     *
     * @param string $method
     * @param string $endPoint
     * @param string|null $body
     * @param array $params
     * @param array $headers
     * @return Daisycon_Affiliatemarketing_Model_Api_Request
     */
    protected function _createRequest(
        $method = Daisycon_Affiliatemarketing_Model_Api_Request::GET,
        $endPoint = '',
        $body = null,
        array $headers = array()
    ) {

        // Add default json headers to the request
        $headers = array_merge(
            $headers,
            array(
                'Accept'       => 'application/json',
                'Content-Type' => 'application/json',
            )
        );

        // Create the request
        return new Daisycon_Affiliatemarketing_Model_Api_Request($method, $endPoint, $headers, $body);
    }

    /**
     * Perform a get request
     *
     * @param string $url
     * @param array $params
     * @param array $headers
     * @return mixed
     */
    public function get($url, array $params = array(), array $headers = array())
    {
        $url = $this->_formatUrl($url, $params);

        try {
            $request = $this->_createAuthenticatedRequest(
                Daisycon_Affiliatemarketing_Model_Api_Request::GET, $url,
                null,
                $params,
                $headers
            );
            $response = $this->_getClient()->send($request);
        } catch (Exception $exception) {
            $this->_parseExceptionForErrorMessages($exception);
            return null;
        }

        // Determine next url
        $headers = $this->_parseResponseHeaders($response);
        if (false === empty($headers['count']) && $headers['count'] > $headers['per_page'] * $headers['page']) {
            $this->_nextUrl = $this->_replaceUrlQueryString($url, array('page' => $headers['page'] + 1));
        } else {
            $this->_nextUrl = null;
        }

        return $this->_parseResponse($response);
    }

    /**
     * Perform a post request
     *
     * @param string $url
     * @param string $body
     * @return mixed
     */
    public function post($url, array $params = array(), $body = '')
    {
        $url = $this->_formatUrl($url, $params);

        try {
            $request  = $this->_createAuthenticatedRequest(
                Daisycon_Affiliatemarketing_Model_Api_Request::POST,
                $url,
                $body
            );
            $response = $this->_getClient()->send($request);
        } catch (Exception $exception) {
            $this->_parseExceptionForErrorMessages($exception);
            return null;
        }

        return $this->_parseResponse($response);
    }

    /**
     * Perform a put request
     *
     * @param string $url
     * @param string $body
     * @return mixed
     */
    public function put($url, array $params = array(), $body = '')
    {
        $url = $this->_formatUrl($url, $params);

        try {
            $request  = $this->_createAuthenticatedRequest('PUT', $url, $body);
            $response = $this->_getClient()->send($request);
        } catch (Exception $exception) {
            $this->_parseExceptionForErrorMessages($exception);
            return null;
        }

        return $this->_parseResponse($response);
    }

    /**
     * Delete a resource
     *
     * @param string $url
     * @param array $params
     * @return mixed
     */
    public function delete($url, array $params = array())
    {
        $url = $this->_formatUrl($url, $params);

        try {
            $request  = $this->_createAuthenticatedRequest('DELETE', $url);
            $response = $this->_getClient()->send($request);
        } catch (Exception $exception) {
            $this->_parseExceptionForErrorMessages($exception);
            return null;
        }

        return $this->_parseResponse($response);
    }

    /**
     * Get the authentication url
     *
     * @return string
     */
    public function getAuthUrl()
    {
        return $this->_authUrl;
    }

    /**
     * @param mixed $accessToken
     */
    public function setAccessToken($accessToken)
    {
        $this->accessToken = $accessToken;
    }

    /**
     * Is authentication required?
     *
     * @return bool
     */
    public function needsAuthentication()
    {
        return true === empty($this->_accessToken);
    }

    /**
     * Get the access token
     *
     * @return string|null
     */
    public function getAccessToken()
    {
        return $this->_accessToken;
    }

    /**
     * Acquire a new access token
     *
     * @throws Daisycon_Affiliatemarketing_Model_Api_InvalidResponseException
     * @throws Mage_Core_Exception
     */
    protected function _acquireAccessToken()
    {
        // If refresh token not yet acquired, do token request
        $body = array(
            'username' => $this->_username,
            'password' => $this->_password,
        );

        $request = $this->_createRequest(
            Daisycon_Affiliatemarketing_Model_Api_Request::POST,
            $this->_formatUrl($this->getAuthUrl()),
            json_encode($body)
        );

        try
        {
            $response = $this->_getClient()->send($request);
        } catch (RuntimeException $exception) {
            throw new Mage_Core_Exception('Unable to not connect');
        }

        if ($response->getStatusCode() <= 300) {
            $this->_accessToken = json_decode($response->getBody());
        } else {
            throw new Daisycon_Affiliatemarketing_Model_Api_InvalidResponseException($response);
        }
    }

    /**
     * Parse the api response and return the body
     *
     * @param Daisycon_Affiliatemarketing_Model_Api_Response $response
     * @return array
     * @throws Mage_Core_Exception
     */
    protected function _parseResponse(Daisycon_Affiliatemarketing_Model_Api_Response $response)
    {
        try {
            if (204 === $response->getStatusCode()) {
                return array();
            }

            $data = json_decode($response->getBody(), true);

            if (400 <= $response->getStatusCode()) {
                $message = true === isset($data['error']) ?
                    $data['error'] :
                    Mage::helper('affiliatemarketing')
                        ->__('Unknown api error, code %1$s', $response->getStatusCode());
                throw new Mage_Core_Exception($message, $response->getStatusCode());
            }

            return $data;
        } catch (RuntimeException $exception) {
            throw new Mage_Core_Exception($exception->getMessage());
        }
    }

    /**
     * Parse the response headers
     *
     * @param Daisycon_Affiliatemarketing_Model_Api_Response $response
     * @return array
     */
    protected function _parseResponseHeaders(Daisycon_Affiliatemarketing_Model_Api_Response $response)
    {
        $headers = $response->getHeaders();

        $parsed = array();

        // Paging
        if (false === empty($headers['X-Total-Count'])) {
            $parsed['count'] = (int)$headers['X-Total-Count'];
            $parsed['per_page'] = (int)$headers['X-Paginator-Per-Page'];
            $parsed['page'] = (int)$headers['X-Paginator-Page'];
        }

        return $parsed;
    }

    /**
     * Format url
     *
     * @param string $endPoint
     * @param array $params
     * @return string
     */
    protected function _formatUrl($endPoint, array $params = array())
    {
        $endPoint = $this->_injectPlaceholders($endPoint, $params);

        return implode('/', array($this->_getApiUrl(), $endPoint));
    }

    /**
     * Replace one or more parameters in the query string
     *
     * @param $url
     * @param array $params
     * @return string
     */
    protected function _replaceUrlQueryString($url, array $params)
    {
        if (true === empty($params)) {
            return $url;
        }

        $parsed = parse_url($url);
        $parsedQuery = array();
        if (true === parse_str(true === isset($parsed['query']) ? $parsed['query'] : '', $parsedQuery)) {
            $parsed['query'] = array_merge(
                $parsedQuery,
                $params
            );
        }

        $url = (true === isset($parsed['scheme']) ? $parsed['scheme'] . '://' : '')
            . (true === isset($parsed['host']) ? $parsed['host'] : '')
            . (true === isset($parsed['path']) ? $parsed['path'] : '')
            . (true === isset($parsed['query']) ? '?' . $parsed['query'] : '');
        return $url;
    }

    /**
     * Inject parameters into the given path
     *
     * @param string $endPoint
     * @param array $params
     * @return string
     */
    protected function _injectPlaceholders($endPoint, array $params)
    {
        $keys = array_keys($params);
        $query = array();
        foreach ($keys as $key) {
            $placeholder = '{' . $key . '}';
            if (false === strpos($endPoint, $placeholder)) {
                $query[$key] = $params[$key];
                continue;
            }

            $endPoint = str_replace($placeholder, $params[$key], $endPoint);
        }

        if (false === empty($query)) {
            $endPoint .= '?' . http_build_query($query);
        }

        return $endPoint;
    }

    /**
     * Parse the response in the Exception to return the error messages
     *
     * @param Exception $exception
     * @throws Mage_Core_Exception
     */
    protected function _parseExceptionForErrorMessages(Exception $exception)
    {
        if (false === $exception instanceof Daisycon_Affiliatemarketing_Model_Api_InvalidResponseException) {
            throw new Mage_Core_Exception($exception->getMessage());
        }

        $response = $exception->getResponse();
        $responseBody = $response->getBody();
        $decodedResponseBody = json_decode($responseBody, true);

        if (null !== $decodedResponseBody && true === isset($decodedResponseBody['error'])) {
            $errorMessage = $decodedResponseBody['error'];
        } else {
            $errorMessage = $responseBody;
        }

        throw new Mage_Core_Exception(
            'Error ' . $response->getStatusCode() .': ' . $errorMessage
        );
    }

    /**
     * Get the base url path
     *
     * @return string
     */
    protected function _getBaseUrl()
    {
        return $this->_baseUrl;
    }

    /**
     * return the full API url
     *
     * @return string
     */
    protected function _getApiUrl()
    {
        return $this->_baseUrl . $this->_apiUrl;
    }

    /**
     * Set base URL for different labels according to
     *
     * @param string $baseUrl
     * @return $this
     */
    public function setBaseUrl($baseUrl)
    {
        $this->_baseUrl = $baseUrl;
        return $this;
    }

    /**
     * Set the api path
     *
     * @param string $apiUrl
     * @return $this
     */
    public function setApiUrl($apiUrl)
    {
        $this->_apiUrl = $apiUrl;
        return $this;
    }

    /**
     * Set the authentication url
     *
     * @param string $authUrl
     */
    public function setAuthUrl($authUrl)
    {
        $this->_authUrl = $authUrl;
    }

    /**
     * Return the next url
     *
     * @return string
     */
    public function getNextUrl()
    {
        return $this->_nextUrl;
    }

    /**
     * Set the username
     *
     * @param $username
     * @return $this
     */
    public function setUsername($username)
    {
        $this->_username = (string)$username;
        return $this;
    }

    /**
     * Get the username
     *
     * @return null|string
     */
    public function getUsername()
    {
        return $this->_username;
    }

    /**
     * Set the password
     *
     * @param $password
     * @return $this
     */
    public function setPassword($password)
    {
        $this->_password = (string)$password;
        return $this;
    }

    /**
     * Get the password
     *
     * @return null|string
     */
    public function getPassword()
    {
        return $this->_password;
    }
}
